/*
 * Copyright (c) 2021, NVIDIA CORPORATION. All rights reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */



#include <cassert>
#include <cfloat>
#include <fstream>
#include <iostream>
#include <memory>
#include <sstream>


#include <cuda_runtime_api.h>
#include "NvInfer.h"
#include "NvOnnxParser.h"  // Delete

#include <opencv2/opencv.hpp>
#include <opencv2/core/core.hpp>
#include <opencv2/highgui/highgui.hpp>


//#include <opencv2/imgcodecs.hpp>
//#include <opencv2/core/cuda.hpp>
#include <opencv2/cudawarping.hpp>
// #include <opencv2/core.hpp>
#include <opencv2/cudaarithm.hpp>


#include "logger.h"
#include "util.h"

constexpr long long operator"" _MiB(long long unsigned val)
{
    return val * (1 << 20);
}



using sample::gLogError;
using sample::gLogInfo;

//!
//! \class SampleSegmentation
//!
//! \brief Implements semantic segmentation using FCN-ResNet101 ONNX model.
//!
class SampleSegmentation
{

public:
    SampleSegmentation(const std::string& engineFilename);
    bool infer(const std::string& input_filename, int32_t width, int32_t height, const std::string& output_filename);

private:
    std::string mEngineFilename;                    //!< Filename of the serialized engine.

    nvinfer1::Dims mInputDims;                      //!< The dimensions of the input to the network.
    nvinfer1::Dims mOutputDims;                     //!< The dimensions of the output to the network.

    util::UniquePtr<nvinfer1::ICudaEngine> mEngine; //!< The TensorRT engine used to run the network
};

SampleSegmentation::SampleSegmentation(const std::string& engineFilename)
    : mEngineFilename(engineFilename)
    , mEngine(nullptr)
{

    std::cout << "Reading File: " << engineFilename << std::endl;
    // De-serialize engine from file
    std::ifstream engineFile(engineFilename, std::ios::binary);
    if (engineFile.fail())
    {

        std::cout << "File Not Found! " << std::endl;
        
        return;
    }

    engineFile.seekg(0, std::ifstream::end);
    auto fsize = engineFile.tellg();
    engineFile.seekg(0, std::ifstream::beg);

    std::vector<char> engineData(fsize);
    engineFile.read(engineData.data(), fsize);

    util::UniquePtr<nvinfer1::IRuntime> runtime{nvinfer1::createInferRuntime(sample::gLogger.getTRTLogger())};
    mEngine.reset(runtime->deserializeCudaEngine(engineData.data(), fsize, nullptr));
    assert(mEngine.get() != nullptr);
}

std::string returnBindingDataType (nvinfer1::DataType type) {

    // DataType : int32_t {kFLOAT = 0 , kHALF = 1 , kINT8 = 2 , kINT32 = 3 , kBOOL = 4 }

    switch (type) {
    case nvinfer1::DataType::kFLOAT:
        return ("kFLOAT"); 
    case nvinfer1::DataType::kHALF:
        return ("kHALF");
    case nvinfer1::DataType::kINT8:
        return ("kINT8");
    case nvinfer1::DataType::kINT32:
        return ("kINT32");
    case nvinfer1::DataType::kBOOL:
        return ("kBOOL");
    }
    return ("UNDEFINED");

}

std::string returnTensorFormat (nvinfer1::TensorFormat thisFormat) {

std::string formatDescription;

    switch (thisFormat) {
        case nvinfer1::TensorFormat::kLINEAR: { formatDescription = "kLINEAR Row major linear format. For a tensor with dimensions {N, C, H, W} or {numbers, channels, columns, rows}, the dimensional index corresponds to {3, 2, 1, 0} and thus the order is W minor. For DLA usage, the tensor sizes are limited to C,H,W in the range [1,8192]. ";
                        break; }
        case nvinfer1::TensorFormat::kCHW2: { formatDescription = "kCHW2 Two wide channel vectorized row major format. This format is bound to FP16. It is only available for dimensions >= 3. For a tensor with dimensions {N, C, H, W}, the memory layout is equivalent to a C array with dimensions [N][(C+1)/2][H][W][2], with the tensor coordiates (n, c, h, w) mapping to array subscript [n][c/2][h][w][c%2]. ";
                        break; }
        case nvinfer1::TensorFormat::kHWC8: { formatDescription = "kHWC8 Eight channel format where C is padded to a multiple of 8. This format is bound to FP16. It is only available for dimensions >= 3. For a tensor with dimensions {N, C, H, W}, the memory layout is equivalent to the array with dimensions [N][H][W][(C+7)/8*8], with the tensor coordinates (n, c, h, w) mapping to array subscript [n][h][w][c]. ";
                        break; }
        case nvinfer1::TensorFormat::kCHW4: { formatDescription = "kCHW4 Four wide channel vectorized row major format. This format is bound to INT8 or FP16. It is only available for dimensions >= 3. For INT8, the C dimension must be a build-time constant. For a tensor with dimensions {N, C, H, W}, the memory layout is equivalent to a C array with dimensions [N][(C+3)/4][H][W][4], with the tensor coordinates (n, c, h, w) mapping to array subscript [n][c/4][h][w][c%4]. Deprecated usage: If running on the DLA, this format can be used for acceleration with the caveat that C must be equal or lesser than 4. If used as DLA input and the build option kGPU_FALLBACK is not specified, it needs to meet line stride requirement of DLA format. Column stride in bytes should be a multiple of 32 on Xavier and 64 on Orin. ";
                        break; }
        case nvinfer1::TensorFormat::kCHW16: { formatDescription = "kCHW16 Sixteen wide channel vectorized row major format. This format is bound to FP16. It is only available for dimensions >= 3. For a tensor with dimensions {N, C, H, W}, the memory layout is equivalent to a C array with dimensions [N][(C+15)/16][H][W][16], with the tensor coordinates (n, c, h, w) mapping to array subscript [n][c/16][h][w][c%16]. For DLA usage, this format maps to the native image format for FP16, and the tensor sizes are limited to C,H,W in the range [1,8192]. ";
                        break; }
        case nvinfer1::TensorFormat::kCHW32: { formatDescription = "kCHW32 Thirty-two wide channel vectorized row major format. This format is only available for dimensions >= 3. For a tensor with dimensions {N, C, H, W}, the memory layout is equivalent to a C array with dimensions [N][(C+31)/32][H][W][32], with the tensor coordinates (n, c, h, w) mapping to array subscript [n][c/32][h][w][c%32]. For DLA usage, this format maps to the native image format for INT8, and the tensor sizes are limited to C,H,W in the range [1,8192]. ";
                        break; }
        case nvinfer1::TensorFormat::kDHWC8: { formatDescription = "kDHWC8 Eight channel format where C is padded to a multiple of 8. This format is bound to FP16, and it is only available for dimensions >= 4. For a tensor with dimensions {N, C, D, H, W}, the memory layout is equivalent to an array with dimensions [N][D][H][W][(C+7)/8*8], with the tensor coordinates (n, c, d, h, w) mapping to array subscript [n][d][h][w][c]. ";
                        break; }
        case nvinfer1::TensorFormat::kCDHW32: { formatDescription = "kCDHW32 Thirty-two wide channel vectorized row major format. This format is bound to FP16 and INT8 and is only available for dimensions >= 4. For a tensor with dimensions {N, C, D, H, W}, the memory layout is equivalent to a C array with dimensions [N][(C+31)/32][D][H][W][32], with the tensor coordinates (n, c, d, h, w) mapping to array subscript [n][c/32][d][h][w][c%32]. ";
                        break; }
        case nvinfer1::TensorFormat::kHWC:  { formatDescription = "kHWC Non-vectorized channel-last format. This format is bound to FP32 and is only available for dimensions >= 3. ";
                        break; }
        case nvinfer1::TensorFormat::kDLA_LINEAR: { formatDescription = "kDLA_LINEAR DLA planar format. For a tensor with dimension {N, C, H, W}, the W axis always has unit stride. The stride for stepping along the H axis is rounded up to 64 bytes. The memory layout is equivalent to a C array with dimensions [N][C][H][roundUp(W, 64/elementSize)] where elementSize is 2 for FP16 and 1 for Int8, with the tensor coordinates (n, c, h, w) mapping to array subscript [n][c][h][w]. ";
                        break; }
        case nvinfer1::TensorFormat::kDLA_HWC4: { formatDescription = "kDLA_HWC4 DLA image format. For a tensor with dimension {N, C, H, W} the C axis always has unit stride. The stride for stepping along the H axis is rounded up to 32 bytes on Xavier and 64 bytes on Orin. C can only be 1, 3 or 4. If C == 1, it will map to grayscale format. If C == 3 or C == 4, it will map to color image format. And if C == 3, the stride for stepping along the W axis needs to be padded to 4 in elements. When C is {1, 3, 4}, then C' is {1, 4, 4} respectively, the memory layout is equivalent to a C array with dimensions [N][H][roundUp(W, 32/C'/elementSize)][C'] where elementSize is 2 for FP16 and 1 for Int8. The tensor coordinates (n, c, h, w) mapping to array subscript [n][h][w][c]. ";
                        break; }
        case nvinfer1::TensorFormat::kHWC16: { formatDescription = "kHWC16 Sixteen channel format where C is padded to a multiple of 16. This format is bound to FP16. It is only available for dimensions >= 3. For a tensor with dimensions {N, C, H, W}, the memory layout is equivalent to the array with dimensions [N][H][W][(C+15)/16*16], with the tensor coordinates (n, c, h, w) mapping to array subscript [n][h][w][c]. ";
                        break; }
        default: { formatDescription = "Undefined"; }

        }


    // std::cout << formatDescription << std::endl;

    return (formatDescription);
    
}

// preprocessing stage ------------------------------------------------------------------------------------------------
void preprocessImage(const std::string& image_path, float* gpu_input, const nvinfer1::Dims& dims)
{
    // read input image
    cv::Mat frame = cv::imread(image_path);
    if (frame.empty())
    {
        std::cerr << "Input image " << image_path << " load failed\n";
        return;
    }
    cv::cuda::GpuMat gpu_frame;
    // upload image to GPU
    gpu_frame.upload(frame);

    auto input_width = dims.d[3];  //{1, 3 /* channels */, height, width};
    auto input_height = dims.d[2];
    auto channels = dims.d[1];
    auto input_size = cv::Size(input_width, input_height);
    // resize
    cv::cuda::GpuMat resized;
    cv::cuda::resize(gpu_frame, resized, input_size, 0, 0, cv::INTER_NEAREST);
    // normalize
    cv::cuda::GpuMat flt_image;
    resized.convertTo(flt_image, CV_32FC3, 1.f / 255.f);
    cv::cuda::subtract(flt_image, cv::Scalar(0.485f, 0.456f, 0.406f), flt_image, cv::noArray(), -1);
    cv::cuda::divide(flt_image, cv::Scalar(0.229f, 0.224f, 0.225f), flt_image, 1, -1);
    // to tensor
    std::vector<cv::cuda::GpuMat> chw;
    for (size_t i = 0; i < channels; ++i)
    {
        chw.emplace_back(cv::cuda::GpuMat(input_size, CV_32FC1, gpu_input + (i * input_width * input_height)));
    }
    cv::cuda::split(flt_image, chw);
}





//!
//! \brief Runs the TensorRT inference.
//!
//! \details Allocate input and output memory, and executes the engine.
//!
bool SampleSegmentation::infer(const std::string& input_filename, int32_t width, int32_t height, const std::string& output_filename)
{
    auto context = util::UniquePtr<nvinfer1::IExecutionContext>(mEngine->createExecutionContext());
    if (!context)
    {
        return false;
    }

    for (int32_t i = 0 ; i < 2; i++) {
        std::cout << "Binding Index " << i << ": " << mEngine->getBindingName(i) << std::endl;
    }
        // Binding Index 0: input_1
        // Binding Index 1: softmax_1

    auto input_idx = mEngine->getBindingIndex("input_1");
    if (input_idx == -1)
    {
        return false;
    }

    std::cout << "TensorFormat for input_1: " << returnTensorFormat (mEngine->getBindingFormat(input_idx)) << std::endl << std::endl;
    std::cout << "input_1 Datatype: " << returnBindingDataType(mEngine->getBindingDataType(input_idx)) << std::endl;
    // TensorFormat for input_1: kLINEAR Row major linear format. For a tensor with dimensions {N, C, H, W} or {numbers, channels, columns, rows}, the dimensional index corresponds to {3, 2, 1, 0} and thus the order is W minor. For DLA usage, the tensor sizes are limited to C,H,W in the range [1,8192]. 
    // input_1 Datatype: kFLOAT


    assert(mEngine->getBindingDataType(input_idx) == nvinfer1::DataType::kFLOAT);
    auto input_dims = nvinfer1::Dims4{1, 3 /* channels */, height, width};
    context->setBindingDimensions(input_idx, input_dims);
    auto myDims = context->getBindingDimensions(input_idx);
    auto input_size = util::getMemorySize(input_dims, sizeof(float));

    // auto output_idx = mEngine->getBindingIndex("output"); 
    auto output_idx = mEngine->getBindingIndex("softmax_1"); // Output layer name in our model
    if (output_idx == -1)
    {
        return false;
    }

    std::cout << "TensorFormat for softmax_1: " << returnTensorFormat (mEngine->getBindingFormat(output_idx)) << std::endl << std::endl;
    std::cout << "softmax_1 Datatype: " << returnBindingDataType(mEngine->getBindingDataType(output_idx)) << std::endl;
    // TensorFormat for softmax_1: kLINEAR Row major linear format. For a tensor with dimensions {N, C, H, W} or {numbers, channels, columns, rows}, the dimensional index corresponds to {3, 2, 1, 0} and thus the order is W minor. For DLA usage, the tensor sizes are limited to C,H,W in the range [1,8192]. 
    // softmax_1 Datatype: kFLOAT

    // assert(mEngine->getBindingDataType(output_idx) == nvinfer1::DataType::kINT32);  DBG TODO
    assert(mEngine->getBindingDataType(output_idx) == nvinfer1::DataType::kFLOAT); 
    auto output_dims = context->getBindingDimensions(output_idx);
    auto output_size = util::getMemorySize(output_dims, sizeof(float));

    // Allocate CUDA memory for input and output bindings
    void* input_mem{nullptr};
    if (cudaMalloc(&input_mem, input_size) != cudaSuccess)
    {
        gLogError << "ERROR: input cuda memory allocation failed, size = " << input_size << " bytes" << std::endl;
        return false;
    }
    void* output_mem{nullptr};
    if (cudaMalloc(&output_mem, output_size) != cudaSuccess)
    {
        gLogError << "ERROR: output cuda memory allocation failed, size = " << output_size << " bytes" << std::endl;
        return false;
    }


    // Read image data from file and mean-normalize it
    const std::vector<float> mean{0.485f, 0.456f, 0.406f};
    const std::vector<float> stddev{0.229f, 0.224f, 0.225f};
    auto input_image{util::RGBImageReader(input_filename, input_dims, mean, stddev)};
    input_image.read();  // Fake call Irrelevant now
    auto input_buffer = input_image.processCV();


    /*******************************
    // Test that input_buffer actualy contains a picture
    // 	Mat (int rows, int cols, int type, void *data, size_t step=AUTO_STEP)
    cv::Mat A (height, width, CV_8UC1, cv::Scalar(0));
    for (int r =0; r< height; r++)
    for (int c =0; c< width; c++)
        A.at<uint8_t>(r, c) = static_cast<uint8_t>(input_buffer.get()[(r*width)+c]);

    int HW = height * width;
    cv::Mat B (height, width, CV_8UC1, cv::Scalar(0));
    for (int r =0; r< height; r++)
    for (int c =0; c< width; c++)
        B.at<uint8_t>(r, c) = static_cast<uint8_t>(input_buffer.get()[HW +(r*width)+c]);

    int HWx2 = height * width *2;
    cv::Mat C (height, width, CV_8UC1, cv::Scalar(0));
    for (int r =0; r< height; r++)
    for (int c =0; c< width; c++)
        C.at<uint8_t>(r, c) = static_cast<uint8_t>(input_buffer.get()[HWx2 +(r*width)+c]);

    cv::Mat merged;
    std::vector<cv::Mat> channels = {A, B, C};
    cv::merge(channels,merged);

    imshow("Tensor Reconstructed",merged);
    cv::waitKey();

    **/

    cudaStream_t stream;
    if (cudaStreamCreate(&stream) != cudaSuccess)
    {
        gLogError << "ERROR: cuda stream creation failed." << std::endl;
        return false;
    }

    // Copy image data to input binding memory
    if (cudaMemcpyAsync(input_mem, input_buffer.get(), input_size, cudaMemcpyHostToDevice, stream) != cudaSuccess)
    {
        gLogError << "ERROR: CUDA memory copy of input failed, size = " << input_size << " bytes" << std::endl;
        return false;
    }

    cudaStreamSynchronize(stream);  // Wait for CUDA to complete



    // Run TensorRT inference
    void* bindings[] = {input_mem, output_mem};
    bool status = context->enqueueV2(bindings, stream, nullptr);
    if (!status)
    {
        gLogError << "ERROR: TensorRT inference failed" << std::endl;
        return false;
    }
    cudaStreamSynchronize(stream);  // Wait for infer to complete






    // Copy predictions from output binding memory
    auto output_buffer = std::unique_ptr<float>{new float[output_size]};
    if (cudaMemcpyAsync(output_buffer.get(), output_mem, output_size, cudaMemcpyDeviceToHost, stream) != cudaSuccess)
    {
        gLogError << "ERROR: CUDA memory copy of output failed, size = " << output_size << " bytes" << std::endl;
        return false;
    }
    cudaStreamSynchronize(stream);  // Wait for infer to complete

    // Plot the semantic segmentation predictions of 21 classes in a colormap image and write to file
    // const int num_classes{21};
    // const std::vector<int> palette{(0x1 << 25) - 1, (0x1 << 15) - 1, (0x1 << 21) - 1};
    // auto output_image{util::ArgmaxImageWriter(output_filename, output_dims, palette, num_classes)};
    // output_image.process(output_buffer.get());
    // output_image.write();

    // Free CUDA resources


        int counts [6] = {0, 0, 0, 0, 0, 0};
        for (int j = 0; j < (height * width); j++) {


              // if ((int)resized.at<Vec3b>(r, c)[2] != 0) cout << (int)resized.at<Vec3b>(r, c)[2] << " " ;
              // switch ((int)red.at<uint8_t>(r, c)) {
              switch (static_cast<uint8_t>(output_buffer.get()[j])) {
              case 0: 
                counts[0]++; break;
              case 1: 
                counts[1]++; break;
              case 2: 
                counts[2]++; break;
              case 3: 
                counts[3]++; break;
              case 4: 
                counts[4]++; break;      
              default:
                counts[5]++; 
                // cout << (int)origFrameMASK.at<Vec3b>(r, c)[0] << ", ";  // These are 255
                break;
              }
            
          }

          // cout << endl << endl;

          std::cout <<  "0: " << counts[0] << ", 1: " << counts[1] << ", 2: " << counts[2] << ", 3: " 
                              << counts[3] << ", 4: " << counts[4] << ", BAD: " << counts[5] << std::endl;



    cv::Mat outMask     (height, width, CV_8UC1);
    cv::Mat displayMask (height, width, CV_8UC3, cv::Scalar(0, 0, 0));

    // for (int row = 0; row < height; row++)
    // for (int col = 0; col < width; col++)
    //     outMask.at<uint>(row, col) = static_cast<uint8_t>(output_buffer.get()[row*width + col]);


    // Update: For Unet, the preprocessing uses CHW order. The postprocessing uses HWC order.
    for (int row = 0; row < height; row++)
    for (int col = 0; col < width; col++)
        // outMask.at<uchar>(row, col) = static_cast<uchar>(output_buffer.get()[col*height + row]);
        outMask.at<uchar>(row, col) = static_cast<uchar>(output_buffer.get()[row*width + col]);

    // cv::imshow("outMask", outMask);
    // cv::waitKey();

    cv::Scalar color;
    for (int i =0; i<= 4; i++) {
    cv::Mat temp;
    cv::inRange(outMask, cv::Scalar(i), cv::Scalar(i), temp);    // Background
    // Mask codes:  End 100,0,0, Stem 70,0,0, Node 40,0,0, Leaf 10,0,0 Background 0
    // Mask codes:  End 1        Stem 1       Node 2       Leaf 1      Background 0

    switch (i) {

        case 0: color = cv::Scalar (255,255,255); 
        //displayMask.setTo(Scalar (255,255,255), temp); break;
        break; // Backround  White
        case 1: color = cv::Scalar (255,0,0);   
        break; // Leaf       Blue
        case 2: color = cv::Scalar (0,76,153);    // (0,76,153)
        break; // Node       Brownish
        case 3: color = cv::Scalar (0,255,0);     
        break; // Stem       Green
        case 4: color = cv::Scalar (0,0,255);     
        break; // End        Red
        default: color = cv::Scalar (0,0,0);      
        break;// Backround
    }
    // displayMask.setTo(Scalar(i*40,i*40,i*40), temp);
    displayMask.setTo(color, temp); 
    }

    cv::imshow("displayMask", displayMask);
    cv::waitKey();


    cv::destroyAllWindows();

    outMask.deallocate();
    displayMask.deallocate();
    cudaFree(input_mem);
    cudaFree(output_mem);
    return true;
}




bool file_exist(const char *fileName)
{
    std::ifstream infile(fileName);
    return infile.good();
}


// calculate size of tensor
size_t getSizeByDim(const nvinfer1::Dims& dims)
{
    size_t size = 1;
    for (size_t i = 0; i < dims.nbDims; ++i)
    {
        size *= dims.d[i];
    }
    return size;
}


int main(int argc, char** argv)
{
    // int32_t width{512};
    // int32_t height{512};
    int32_t width{960};
    int32_t height{544};

    // SampleSegmentation sample("/home/david/Envs/TAO/6S001/export/int8.tlt.engine");
    // SampleSegmentation sample("/home/david/Envs/TAO/6S001/export/6S001_fp16.engine");
    SampleSegmentation sample("/home/david/Downloads/PeopleSemSegNet/bs1_fp16.engine");  Here // Also change in processCV 

    // file_exist("../input.png") ? std::cout << "../input.png Exists\n" : std::cout << "../input.png Doesn't exist\n";
    file_exist("/home/david/Downloads/input_11ft45deg_000070.jpg") ? std::cout << "Input file Exists\n" : std::cout << "Input file doesn't exist\n";
    gLogInfo << "Running TensorRT inference for bs1_fp16.engine" << std::endl;

    
    // if (!sample.infer("../0009.png", width, height, "../output.ppm"))
    if (!sample.infer("/home/david/Downloads/input_11ft45deg_000070.jpg", width, height, "../output.ppm")) Here
    {
        return -1;
    }

    return 0;
}
